"""
Density Functional Theory Utilities
Derived from: https://github.com/mfkasim1/xcnn/blob/f2cb9777da2961ac553f256ecdcca3e314a538ca/xcdnn2/kscalc.py """
import hashlib
import warnings
from dataclasses import dataclass
from abc import abstractmethod, abstractproperty
from typing import Union, List, TypeVar, Generic, Callable
try:
    import torch
except Exception as e:
    warnings.warn("Could not import torch. Skipping tests." + str(e))

try:
    import xitorch as xt
except Exception as e:
    warnings.warn("Could not import xitorch. Skipping tests." + str(e))


T = TypeVar('T')


@dataclass
class SpinParam(Generic[T]):
    """
    Data structure to store different values for spin-up and spin-down electrons.

    Attributes
    ----------
    u: any type
        The parameters that corresponds to the spin-up electrons.
    d: any type
        The parameters that corresponds to the spin-down electrons.

    References
    ----------
    Kasim, Muhammad F., and Sam M. Vinko. "Learning the exchange-correlation
    functional from nature with fully differentiable density functional
    theory." Physical Review Letters 127.12 (2021): 126403.
    https://github.com/diffqc/dqc/blob/master/dqc/utils/datastruct.py
    """

    u: T
    d: T

    def sum(self):
        """
        Returns the sum of up and down parameters
        """

        return self.u + self.d

    def reduce(self, fcn: Callable) -> T:
        """
        Reduce up and down parameters with the given function
        """

        return fcn(self.u, self.d)


class BaseQCCalc(object):
    """
    Quantum Chemistry calculation. This class is the interface to the users
    regarding parameters that can be calculated after the self-consistent
    iterations (or other processes).

    References
    ----------
    Kasim, Muhammad F., and Sam M. Vinko. "Learning the exchange-correlation
    functional from nature with fully differentiable density functional
    theory." Physical Review Letters 127.12 (2021): 126403.
    https://github.com/diffqc/dqc/blob/master/dqc/utils/datastruct.py
    """

    @abstractmethod
    def get_system(self):
        """
        Returns the system in the QC calculation
        """
        pass

    @abstractmethod
    def run(self, **kwargs):
        """
        Run the calculation.
        Note that this method can be invoked several times for one object to
        try for various self-consistent options to reach convergence.
        """
        pass

    @abstractmethod
    def energy(self) -> torch.Tensor:
        """
        Obtain the energy of the system.
        """
        pass

    @abstractmethod
    def aodm(self) -> Union[torch.Tensor, SpinParam[torch.Tensor]]:
        """
        Returns the density matrix in atomic orbital. For polarized case, it
        returns a SpinParam of 2 tensors representing the density matrices for
        spin-up and spin-down.
        """
        # return: (nao, nao)
        pass

    @abstractmethod
    def dm2energy(
            self, dm: Union[torch.Tensor,
                            SpinParam[torch.Tensor]]) -> torch.Tensor:
        """
        Calculate the energy from the given density matrix.

        Arguments
        ---------
        dm: torch.Tensor or SpinParam of torch.Tensor
            The input density matrix. It is tensor if restricted, and SpinParam
            of tensor if unrestricted.

        Returns
        -------
        torch.Tensor
            Tensor that represents the energy given the energy.
        """
        pass

    @abstractmethod
    def getparamnames(self, methodname: str, prefix: str = "") -> List[str]:
        """
        Return a list with the parameter names corresponding to the given method
        (methodname)

        Returns
        -------
        List[str]
            List of parameter names of methodname

        """

        pass


class KSCalc(object):
    """
    Interface to DQC's KS calculation.

    Parameters
    ----------
    qc: BaseQCCalc
        object often acts as a wrapper around an engine class (from dqc.qccalc) that contains information about the self-consistent iterations.

    References
    ----------
    Kasim, Muhammad F., and Sam M. Vinko. "Learning the exchange-correlation functional from nature with fully differentiable density functional theory." Physical Review Letters 127.12 (2021): 126403.
    https://github.com/diffqc/dqc/blob/master/dqc/qccalc/ks.py
    """

    def __init__(self, qc: "BaseQCCalc"):
        self.qc = qc

    def energy(self) -> torch.Tensor:
        """
        Returns
        -------
        The total energy of the Kohn-Sham calculation for a particular system.
        """
        return self.qc.energy()

    def aodmtot(self) -> torch.Tensor:
        """
        Both interacting and non-interacting system's total energy can be expressed in terms of the density matrix. The ground state properties of a system can be calculated by minimizing the energy w.r.t the density matrix.

        Returns
        -------
        The total density matrix in atomic orbital bases.
        """
        dm = self.qc.aodm()
        if isinstance(dm, SpinParam):
            dmtot = dm.u + dm.d
        else:
            dmtot = dm
        return dmtot

    def dens(self, rgrid: torch.Tensor) -> torch.Tensor:
        """
        The ground state density n(r) of a system.

        Parameters
        ----------
        rgrid: torch.Tensor
            Calculate integration grid using dqc.grid.

        Returns
        -------
        The total density profile in the given grid

        Reference
        ---------
        https://github.com/diffqc/dqc/blob/master/dqc/grid/base_grid.py
        """
        dmtot = self.aodmtot()
        return self.qc.get_system().get_hamiltonian().aodm2dens(dmtot, rgrid)

    def force(self) -> torch.Tensor:
        """
        The force on an atom is calculated as the gradient of energy with respect to the atomic position.

        Returns
        -------
        The force for each atom.
        """
        ene = self.energy()
        atompos = self.qc.get_system().atompos
        is_grad_enabled = torch.is_grad_enabled()
        f, = torch.autograd.grad(ene,
                                 atompos,
                                 create_graph=is_grad_enabled,
                                 retain_graph=True)
        return f


def hashstr(s: str) -> str:
    """
    Encodes the string into hashed format - hexadecimal digits.

    Parameters
    ----------
    s : str
    """
    return str(hashlib.blake2s(str.encode(s)).hexdigest())


class BaseGrid(xt.EditableModule):
    """
    Interface to DQC's BaseGrid class. BaseGrid is a class that regulates the integration points over the spatial
    dimensions.
    Parameters
    ----------
    qc: BaseQCCalc
        object often acts as a wrapper around an engine class (from dqc.qccalc) that contains information about the self-consistent iterations.
    References
    ----------
    Kasim, Muhammad F., and Sam M. Vinko. "Learning the exchange-correlation functional from nature with fully differentiable density functional theory." Physical Review Letters 127.12 (2021): 126403.
    https://github.com/diffqc/dqc/blob/0fe821fc92cb3457fb14f6dff0c223641c514ddb/dqc/grid/base_grid.py
    """

    @abstractproperty
    def dtype(self) -> torch.dtype:
        pass

    @abstractproperty
    def device(self) -> torch.device:
        pass

    @abstractproperty
    def coord_type(self) -> str:
        """
        Returns the type of the coordinate returned in get_rgrid
        """
        pass

    @abstractmethod
    def get_dvolume(self) -> torch.Tensor:
        """
        Obtain the torch.tensor containing the dV elements for the integration.
        Returns
        -------
        torch.tensor (*BG, ngrid)
            The dV elements for the integration. *BG is the length of the BaseGrid.
        """
        pass

    @abstractmethod
    def get_rgrid(self) -> torch.Tensor:
        """
        Returns the grid points position in the specified coordinate in
        self.coord_type.
        Returns
        -------
        torch.tensor (*BG, ngrid, ndim)
            The grid points position. *BG is the length of the BaseGrid.
        """
        pass

    @abstractmethod
    def getparamnames(self, methodname: str, prefix: str = "") -> List[str]:
        """
        Return a list with the parameter names corresponding to the given method
        (methodname)
        Returns
        -------
        List[str]
            List of parameter names of methodname
        """
        pass



"""
Simple utils to save and load from disk.
"""
import joblib
import gzip
import pickle
import os
import tempfile
import tarfile
import zipfile
import logging
from urllib.request import urlretrieve
from typing import Any, Iterator, List, Optional, Tuple, Union, cast, IO

import pandas as pd
import numpy as np

import deepchem as dc

logger = logging.getLogger(__name__)


def pad_array(x: np.ndarray,
              shape: Union[Tuple, int],
              fill: float = 0.0,
              both: bool = False) -> np.ndarray:
    """
    Pad an array with a fill value.

    Parameters
    ----------
    x: np.ndarray
        A numpy array.
    shape: Tuple or int
        Desired shape. If int, all dimensions are padded to that size.
    fill: float, optional (default 0.0)
        The padded value.
    both: bool, optional (default False)
        If True, split the padding on both sides of each axis. If False,
        padding is applied to the end of each axis.

    Returns
    -------
    np.ndarray
        A padded numpy array
    """
    x = np.asarray(x)
    if not isinstance(shape, tuple):
        shape = tuple(shape for _ in range(x.ndim))
    pad = []
    for i in range(x.ndim):
        diff = shape[i] - x.shape[i]
        assert diff >= 0
        if both:
            a, b = divmod(diff, 2)
            b += a
            pad.append((a, b))
        else:
            pad.append((0, diff))
    pad = tuple(pad)  # type: ignore
    x = np.pad(x, pad, mode='constant', constant_values=fill)
    return x


def get_data_dir() -> str:
    """Get the DeepChem data directory.

    Returns
    -------
    str
        The default path to store DeepChem data. If you want to
        change this path, please set your own path to `DEEPCHEM_DATA_DIR`
        as an environment variable.
    """
    if 'DEEPCHEM_DATA_DIR' in os.environ:
        return os.environ['DEEPCHEM_DATA_DIR']
    return tempfile.gettempdir()


def download_url(url: str,
                 dest_dir: str = get_data_dir(),
                 name: Optional[str] = None):
    """Download a file to disk.

    Parameters
    ----------
    url: str
        The URL to download from
    dest_dir: str
        The directory to save the file in
    name: str
        The file name to save it as.  If omitted, it will try to extract a file name from the URL
    """
    if name is None:
        name = url
        if '?' in name:
            name = name[:name.find('?')]
        if '/' in name:
            name = name[name.rfind('/') + 1:]
    if not os.path.exists(dest_dir):
        os.makedirs(dest_dir)
    urlretrieve(url, os.path.join(dest_dir, name))


def untargz_file(file: str,
                 dest_dir: str = get_data_dir(),
                 name: Optional[str] = None):
    """Untar and unzip a .tar.gz file to disk.

    Parameters
    ----------
    file: str
        The filepath to decompress
    dest_dir: str
        The directory to save the file in
    name: str
        The file name to save it as.  If omitted, it will use the file name
    """
    if name is None:
        name = file
    tar = tarfile.open(name)
    tar.extractall(path=dest_dir)
    tar.close()


def unzip_file(file: str,
               dest_dir: str = get_data_dir(),
               name: Optional[str] = None):
    """Unzip a .zip file to disk.

    Parameters
    ----------
    file: str
        The filepath to decompress
    dest_dir: str
        The directory to save the file in
    name: str
        The directory name to unzip it to.  If omitted, it will use the file name
    """
    if name is None:
        name = file
    if dest_dir is None:
        dest_dir = os.path.join(get_data_dir, name)
    with zipfile.ZipFile(file, "r") as zip_ref:
        zip_ref.extractall(dest_dir)


class UniversalNamedTemporaryFile:
    """The implementation for cross platform NamedTemporaryFile.

    `tempfile.NamedTemporaryFile` causes a permission error on Windows.
    This implementation avoids the error, please see threads on the stackoverflow [1]_.

    References
    ----------
    .. [1] https://stackoverflow.com/questions/23212435/permission-denied-to-write-to-my-temporary-file
    """

    def __init__(self, mode='wb', delete=True):
        self._mode = mode
        self._delete = delete

    def __enter__(self):
        # Generate a random temporary file name
        file_name = os.path.join(tempfile.gettempdir(), os.urandom(24).hex())
        # Ensure the file is created
        open(file_name, "x").close()
        # Open the file in the given mode
        self._tempFile = open(file_name, self._mode)
        return self._tempFile

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._tempFile.close()
        if self._delete:
            os.remove(self._tempFile.name)


def load_image_files(input_files: List[str]) -> np.ndarray:
    """Loads a set of images from disk.

    Parameters
    ----------
    input_files: List[str]
        List of image filenames.

    Returns
    -------
    np.ndarray
        A numpy array that contains loaded images. The shape is, `(N,...)`.

    Notes
    -----
    This method requires Pillow to be installed.
    The supported file types are PNG and TIF.
    """
    try:
        from PIL import Image
    except ModuleNotFoundError:
        raise ImportError("This function requires Pillow to be installed.")

    images = []
    for input_file in input_files:
        _, extension = os.path.splitext(input_file)
        extension = extension.lower()
        if extension == ".png":
            image = np.array(Image.open(input_file))
            images.append(image)
        elif extension == ".tif":
            im = Image.open(input_file)
            imarray = np.array(im)
            images.append(imarray)
        else:
            raise ValueError("Unsupported image filetype for %s" % input_file)
    # dtype=object allows for arrays(images here) of arbitrary size
    try:
        return np.array(images)
    except:
        return np.array(images, dtype=object)


def load_sdf_files(input_files: List[str],
                   clean_mols: bool = True,
                   tasks: List[str] = [],
                   shard_size: Optional[int] = None) -> Iterator[pd.DataFrame]:
    """Load SDF file into dataframe.

    Parameters
    ----------
    input_files: List[str]
        List of filenames
    clean_mols: bool, default True
        Whether to sanitize molecules.
    tasks: List[str], default []
        Each entry in `tasks` is treated as a property in the SDF file and is
        retrieved with `mol.GetProp(str(task))` where `mol` is the RDKit mol
        loaded from a given SDF entry.
    shard_size: int, default None
        The shard size to yield at one time.

    Returns
    -------
    Iterator[pd.DataFrame]
        Generator which yields the dataframe which is the same shard size.

    Notes
    -----
    This function requires RDKit to be installed.
    """
    try:
        from rdkit import Chem
    except ModuleNotFoundError:
        raise ImportError("This function requires RDKit to be installed.")

    df_rows = []
    for input_file in input_files:
        # Tasks are either in .sdf.csv file or in the .sdf file itself for QM9 dataset
        has_csv = os.path.isfile(input_file + ".csv")
        # Structures are stored in .sdf file
        logger.info("Reading structures from %s." % input_file)

        suppl = Chem.SDMolSupplier(str(input_file),
                                   sanitize=clean_mols,
                                   removeHs=False,
                                   strictParsing=False)
        for ind, mol in enumerate(suppl):
            if mol is None:
                continue
            smiles = Chem.MolToSmiles(mol)
            df_row = [ind, smiles, mol]
            if not has_csv:  # Get task targets from .sdf file
                for task in tasks:
                    df_row.append(mol.GetProp(str(task)))

            conf = mol.GetConformer()
            positions = conf.GetPositions()
            pos_x, pos_y, pos_z = zip(*positions)
            df_row.append(str(pos_x))
            df_row.append(str(pos_y))
            df_row.append(str(pos_z))
            df_rows.append(df_row)

            if shard_size is not None and len(df_rows) == shard_size:
                if has_csv:
                    mol_df = pd.DataFrame(df_rows,
                                          columns=('mol_id', 'smiles', 'mol',
                                                   'pos_x', 'pos_y', 'pos_z'))
                    raw_df = next(
                        load_csv_files([input_file + ".csv"], shard_size=None))
                    yield pd.concat([mol_df, raw_df], axis=1, join='inner')
                else:
                    # Note: Here, the order of columns is based on the order in which the values
                    # are appended to `df_row`. Since pos_x, pos_y, pos_z are appended after appending
                    # tasks above, they occur after `tasks` here.
                    # FIXME Ideally, we should use something like a dictionary here to keep it independent
                    # of column ordering.
                    mol_df = pd.DataFrame(df_rows,
                                          columns=('mol_id', 'smiles', 'mol') +
                                          tuple(tasks) +
                                          ('pos_x', 'pos_y', 'pos_z'))
                    yield mol_df
                # Reset aggregator
                df_rows = []

        # Handle final leftovers for this file
        if len(df_rows) > 0:
            if has_csv:
                mol_df = pd.DataFrame(df_rows,
                                      columns=('mol_id', 'smiles', 'mol',
                                               'pos_x', 'pos_y', 'pos_z'))
                raw_df = next(
                    load_csv_files([input_file + ".csv"], shard_size=None))
                yield pd.concat([mol_df, raw_df], axis=1, join='inner')
            else:
                mol_df = pd.DataFrame(df_rows,
                                      columns=('mol_id', 'smiles', 'mol') +
                                      tuple(tasks) +
                                      ('pos_x', 'pos_y', 'pos_z'))
                yield mol_df
            df_rows = []


def load_csv_files(input_files: List[str],
                   shard_size: Optional[int] = None) -> Iterator[pd.DataFrame]:
    """Load data as pandas dataframe from CSV files.

    Parameters
    ----------
    input_files: List[str]
        List of filenames
    shard_size: int, default None
        The shard size to yield at one time.

    Returns
    -------
    Iterator[pd.DataFrame]
        Generator which yields the dataframe which is the same shard size.
    """
    # First line of user-specified CSV *must* be header.
    shard_num = 1
    for input_file in input_files:
        if shard_size is None:
            yield pd.read_csv(input_file)
        else:
            logger.info("About to start loading CSV from %s" % input_file)
            for df in pd.read_csv(input_file, chunksize=shard_size):
                logger.info("Loading shard %d of size %s." %
                            (shard_num, str(shard_size)))
                df = df.replace(np.nan, str(""), regex=True)
                shard_num += 1
                yield df


def load_json_files(input_files: List[str],
                    shard_size: Optional[int] = None) -> Iterator[pd.DataFrame]:
    """Load data as pandas dataframe.

    Parameters
    ----------
    input_files: List[str]
        List of json filenames.
    shard_size: int, default None
        Chunksize for reading json files.

    Returns
    -------
    Iterator[pd.DataFrame]
        Generator which yields the dataframe which is the same shard size.

    Notes
    -----
    To load shards from a json file into a Pandas dataframe, the file
    must be originally saved with ``df.to_json('filename.json', orient='records', lines=True)``
    """
    shard_num = 1
    for input_file in input_files:
        if shard_size is None:
            yield pd.read_json(input_file, orient='records', lines=True)
        else:
            logger.info("About to start loading json from %s." % input_file)
            for df in pd.read_json(input_file,
                                   orient='records',
                                   chunksize=shard_size,
                                   lines=True):
                logger.info("Loading shard %d of size %s." %
                            (shard_num, str(shard_size)))
                df = df.replace(np.nan, str(""), regex=True)
                shard_num += 1
                yield df


def load_pickle_file(input_file: str) -> Any:
    """Load from single, possibly gzipped, pickle file.

    Parameters
    ----------
    input_file: str
        The filename of pickle file. This function can load from
        gzipped pickle file like `XXXX.pkl.gz`.

    Returns
    -------
    Any
        The object which is loaded from the pickle file.
    """
    if ".gz" in input_file:
        with gzip.open(input_file, "rb") as unzipped_file:
            return pickle.load(cast(IO[bytes], unzipped_file))
    else:
        with open(input_file, "rb") as opened_file:
            return pickle.load(opened_file)


def load_pickle_files(input_files: List[str]) -> Iterator[Any]:
    """Load dataset from pickle files.

    Parameters
    ----------
    input_files: List[str]
        The list of filenames of pickle file. This function can load from
        gzipped pickle file like `XXXX.pkl.gz`.

    Returns
    -------
    Iterator[Any]
        Generator which yields the objects which is loaded from each pickle file.
    """
    for input_file in input_files:
        yield load_pickle_file(input_file)


def load_data(input_files: List[str],
              shard_size: Optional[int] = None) -> Iterator[Any]:
    """Loads data from files.

    Parameters
    ----------
    input_files: List[str]
        List of filenames.
    shard_size: int, default None
        Size of shard to yield

    Returns
    -------
    Iterator[Any]
        Iterator which iterates over provided files.

    Notes
    -----
    The supported file types are SDF, CSV and Pickle.
    """
    if len(input_files) == 0:
        raise ValueError("The length of `filenames` must be more than 1.")

    file_type = _get_file_type(input_files[0])
    if file_type == "sdf":
        if shard_size is not None:
            logger.info("Ignoring shard_size for sdf input.")
        for value in load_sdf_files(input_files):
            yield value
    elif file_type == "csv":
        for value in load_csv_files(input_files, shard_size):
            yield value
    elif file_type == "pickle":
        if shard_size is not None:
            logger.info("Ignoring shard_size for pickle input.")
        for value in load_pickle_files(input_files):
            yield value


def _get_file_type(input_file: str) -> str:
    """Get type of input file. Must be csv/pkl/sdf/joblib file."""
    filename, file_extension = os.path.splitext(input_file)
    # If gzipped, need to compute extension again
    if file_extension == ".gz":
        filename, file_extension = os.path.splitext(filename)
    if file_extension == ".csv":
        return "csv"
    elif file_extension == ".pkl":
        return "pickle"
    elif file_extension == ".joblib":
        return "joblib"
    elif file_extension == ".sdf":
        return "sdf"
    else:
        raise ValueError("Unrecognized extension %s" % file_extension)


def save_to_disk(dataset: Any, filename: str, compress: int = 3):
    """Save a dataset to file.

    Parameters
    ----------
    dataset: str
        A data saved
    filename: str
        Path to save data.
    compress: int, default 3
        The compress option when dumping joblib file.
    """
    if filename.endswith('.joblib'):
        joblib.dump(dataset, filename, compress=compress)
    elif filename.endswith('.npy'):
        np.save(filename, dataset)
    else:
        raise ValueError("Filename with unsupported extension: %s" % filename)


def load_from_disk(filename: str) -> Any:
    """Load a dataset from file.

    Parameters
    ----------
    filename: str
        A filename you want to load data.

    Returns
    -------
    Any
        A loaded object from file.
    """
    name = filename
    if os.path.splitext(name)[1] == ".gz":
        name = os.path.splitext(name)[0]
    extension = os.path.splitext(name)[1]
    if extension == ".pkl":
        return load_pickle_file(filename)
    elif extension == ".joblib":
        return joblib.load(filename)
    elif extension == ".csv":
        # First line of user-specified CSV *must* be header.
        df = pd.read_csv(filename, header=0)
        df = df.replace(np.nan, str(""), regex=True)
        return df
    elif extension == ".npy":
        return np.load(filename, allow_pickle=True)
    else:
        raise ValueError("Unrecognized filetype for %s" % filename)


def load_dataset_from_disk(
    save_dir: str
) -> Tuple[bool, Optional[Tuple["dc.data.DiskDataset", "dc.data.DiskDataset",
                                "dc.data.DiskDataset"]],
           List["dc.trans.Transformer"]]:
    """Loads MoleculeNet train/valid/test/transformers from disk.

    Expects that data was saved using `save_dataset_to_disk` below. Expects the
    following directory structure for `save_dir`:
    save_dir/
      |
      ---> train_dir/
      |
      ---> valid_dir/
      |
      ---> test_dir/
      |
      ---> transformers.pkl

    Parameters
    ----------
    save_dir: str
        Directory name to load datasets.

    Returns
    -------
    loaded: bool
        Whether the load succeeded
    all_dataset: Tuple[DiskDataset, DiskDataset, DiskDataset]
        The train, valid, test datasets
    transformers: Transformer
        The transformers used for this dataset

    See Also
    --------
    save_dataset_to_disk
    """

    train_dir = os.path.join(save_dir, "train_dir")
    valid_dir = os.path.join(save_dir, "valid_dir")
    test_dir = os.path.join(save_dir, "test_dir")
    if not os.path.exists(train_dir) or not os.path.exists(
            valid_dir) or not os.path.exists(test_dir):
        return False, None, list()
    loaded = True
    train = dc.data.DiskDataset(train_dir)
    valid = dc.data.DiskDataset(valid_dir)
    test = dc.data.DiskDataset(test_dir)
    train.memory_cache_size = 40 * (1 << 20)  # 40 MB
    all_dataset = (train, valid, test)
    transformers = load_transformers(save_dir)
    return loaded, all_dataset, transformers


def save_dataset_to_disk(save_dir: str, train: "dc.data.DiskDataset",
                         valid: "dc.data.DiskDataset",
                         test: "dc.data.DiskDataset",
                         transformers: List["dc.trans.Transformer"]):
    """Utility used by MoleculeNet to save train/valid/test datasets.

    This utility function saves a train/valid/test split of a dataset along
    with transformers in the same directory. The saved datasets will take the
    following structure:
    save_dir/
      |
      ---> train_dir/
      |
      ---> valid_dir/
      |
      ---> test_dir/
      |
      ---> transformers.pkl

    Parameters
    ----------
    save_dir: str
        Directory name to save datasets to.
    train: DiskDataset
        Training dataset to save.
    valid: DiskDataset
        Validation dataset to save.
    test: DiskDataset
        Test dataset to save.
    transformers: List[Transformer]
        List of transformers to save to disk.

    See Also
    --------
    load_dataset_from_disk
    """
    train_dir = os.path.join(save_dir, "train_dir")
    valid_dir = os.path.join(save_dir, "valid_dir")
    test_dir = os.path.join(save_dir, "test_dir")
    train.move(train_dir)
    valid.move(valid_dir)
    test.move(test_dir)
    save_transformers(save_dir, transformers)


def load_transformers(save_dir: str) -> List["dc.trans.Transformer"]:
    """Load the transformers for a MoleculeNet dataset from disk."""
    with open(os.path.join(save_dir, "transformers.pkl"), 'rb') as f:
        return pickle.load(f)


def save_transformers(save_dir: str,
                      transformers: List["dc.trans.Transformer"]):
    """Save the transformers for a MoleculeNet dataset to disk."""
    with open(os.path.join(save_dir, "transformers.pkl"), 'wb') as f:
        pickle.dump(transformers, f)


def remove_missing_entries(dataset):
    """Remove missing entries.

    Some of the datasets have missing entries that sneak in as zero'd out
    feature vectors. Get rid of them.
    """
    for i, (X, y, w, ids) in enumerate(dataset.itershards()):
        available_rows = X.any(axis=1)
        logger.info("Shard %d has %d missing entries." %
                    (i, np.count_nonzero(~available_rows)))
        X = X[available_rows]
        y = y[available_rows]
        w = w[available_rows]
        ids = ids[available_rows]
        dataset.set_shard(i, X, y, w, ids)



import torch
from abc import abstractmethod
from typing import List, Callable, Union
from deepchem.utils.dft_utils import RadialGrid


class BaseTruncationRules(object):
    """Base class to store the truncation rules of an individual atomic grid.

    Examples
    --------
    >>> from deepchem.utils.dft_utils import BaseTruncationRules
    >>> class MyTrunc(BaseTruncationRules):
    ...     def to_truncate(self, atm: int) -> bool:
    ...         return False
    >>> trunc = MyTrunc()
    >>> trunc.to_truncate(1)
    False

    """

    @abstractmethod
    def to_truncate(self, atz: int) -> bool:
        """Decide whether to truncate the atom's grid

        Parameters
        ----------
        atz: int
            Atomic number of the atom

        Returns
        -------
        bool
            True if the grid should be truncated, False otherwise

        """
        pass

    @abstractmethod
    def rad_slices(self, atz: int, radgrid: RadialGrid) -> List[slice]:
        """Get the list of slices of radial grid

        Parameters
        ----------
        atz: int
            Atomic number of the atom
        radgrid: RadialGrid
            RadialGrid object of the atom

        Returns
        -------
        List[slice]
            List of slices of the radial grid

        """
        pass

    @abstractmethod
    def precs(self, atz: int, radgrid: RadialGrid) -> List[int]:
        """Get the list of precisions of angular grid for each slice in the
        sliced radial grids

        Parameters
        ----------
        atz: int
            Atomic number of the atom
        radgrid: RadialGrid
            RadialGrid object of the atom

        """
        pass


class NoTrunc(BaseTruncationRules):
    """No truncation rule. Use the full grid for all atoms.

    Examples
    --------
    >>> from deepchem.utils.dft_utils import NoTrunc
    >>> rule = NoTrunc()
    >>> rule.to_truncate(1)
    False

    """

    def __init__(self):
        """Initialize the NoTrunc object"""
        pass

    def to_truncate(self, atz: int) -> bool:
        """Decide whether to truncate the atom's grid

        Parameters
        ----------
        atz: int
            Atomic number of the atom

        Returns
        -------
        bool
            Always False

        """
        return False

    def rad_slices(self, atz: int, radgrid: RadialGrid) -> List[slice]:
        """Get the list of slices of radial grid

        Parameters
        ----------
        atz: int
            Atomic number of the atom
        radgrid: RadialGrid
            RadialGrid object of the atom

        Returns
        -------
        List[slice]
            Always raises RuntimeError

        """
        raise RuntimeError("This shouldn't be called. Report to Github")

    def precs(self, atz: int, radgrid: RadialGrid) -> List[int]:
        """Get the list of precisions of angular grid for each slice in the
        sliced radial grids

        Parameters
        ----------
        atz: int
            Atomic number of the atom
        radgrid: RadialGrid
            RadialGrid object of the atom

        Returns
        -------
        List[int]
            Always raises RuntimeError

        """
        raise RuntimeError("This shouldn't be called. Report to Github")


class DasguptaTrunc(BaseTruncationRules):
    """
    Truncation rule from Dasgupta et al., https://onlinelibrary.wiley.com/doi/epdf/10.1002/jcc.24761

    Examples
    --------
    >>> from deepchem.utils.dft_utils import DasguptaTrunc
    >>> rule = DasguptaTrunc(75)
    >>> rule.to_truncate(1)
    True

    """

    def __init__(self, nr: Union[int, Callable[[int], int]]):
        """Initialize the DasguptaTrunc object.

        Parameters
        ----------
        nr: Union[int, Callable[[int], int]]
            Number of radial points or a function of atomic number to
            get the number of radial points.

        """
        self._truncate_idxs = {
            75: {
                1: [0, 35, 47, 63, 70, 75],
                3: [0, 35, 47, 64, 71, 75],
                4: [0, 35, 47, 64, 71, 75],
                5: [0, 35, 47, 64, 71, 75],
                6: [0, 35, 47, 64, 71, 75],
                7: [0, 35, 47, 64, 71, 75],
                8: [0, 30, 44, 62, 70, 75],
                9: [0, 26, 42, 61, 69, 75],
                11: [0, 35, 47, 64, 71, 75],
                12: [0, 35, 47, 64, 71, 75],
                13: [0, 32, 47, 64, 71, 75],
                14: [0, 32, 47, 64, 71, 75],
                15: [0, 30, 44, 61, 68, 75],
                16: [0, 30, 44, 61, 68, 75],
                17: [0, 26, 42, 61, 69, 75],
            },
            99: {
                1: [0, 45, 61, 82, 92, 99],
                3: [0, 46, 62, 84, 93, 99],
                4: [0, 42, 48, 62, 84, 87, 93, 99],
                5: [0, 42, 48, 62, 84, 93, 99],
                6: [0, 46, 62, 84, 85, 87, 93, 99],
                7: [0, 40, 58, 82, 93, 99],
                8: [0, 40, 54, 56, 58, 82, 83, 84, 92, 99],
                9: [0, 35, 52, 56, 81, 83, 91, 99],
                11: [0, 46, 62, 84, 93, 99],
                12: [0, 48, 63, 83, 90, 99],
                13: [0, 42, 48, 62, 84, 87, 93, 99],
                14: [0, 42, 48, 62, 84, 93, 99],
                15: [0, 35, 36, 54, 58, 83, 85, 93, 99],
                16: [0, 35, 36, 54, 58, 83, 85, 93, 99],
                17: [0, 35, 52, 56, 81, 83, 91, 99],
            },
        }
        self._truncate_precs = {
            75: {
                1: [3, 17, 29, 15, 7],
                3: [3, 17, 29, 15, 11],
                4: [3, 17, 29, 15, 11],
                5: [3, 17, 29, 19, 7],
                6: [3, 17, 29, 19, 7],
                7: [3, 17, 29, 15, 7],
                8: [3, 17, 29, 19, 11],
                9: [3, 17, 29, 17, 11],
                11: [3, 17, 29, 15, 11],
                12: [3, 17, 29, 15, 11],
                13: [3, 17, 29, 19, 11],
                14: [3, 17, 29, 19, 11],
                15: [3, 17, 29, 19, 9],
                16: [3, 17, 29, 19, 9],
                17: [3, 17, 29, 17, 11],
            },
            99: {
                1: [3, 17, 41, 23, 11],
                3: [3, 17, 41, 19, 11],
                4: [3, 15, 17, 41, 23, 19, 11],
                5: [3, 15, 17, 41, 23, 11],
                6: [3, 19, 41, 29, 23, 19, 15],
                7: [3, 17, 41, 19, 11],
                8: [3, 17, 23, 29, 41, 29, 23, 19, 11],
                9: [3, 17, 23, 41, 23, 17, 11],
                11: [3, 17, 41, 19, 11],
                12: [3, 17, 41, 19, 11],
                13: [3, 15, 17, 41, 23, 19, 11],
                14: [3, 15, 17, 41, 23, 11],
                15: [3, 15, 17, 23, 41, 23, 19, 11],
                16: [3, 15, 17, 23, 41, 23, 19, 11],
                17: [3, 17, 23, 41, 23, 17, 11],
            },
        }
        self._nr = nr

    def _get_truncate_idxs(self, atz: int) -> List[int]:
        """Return the truncate indices for the given atom z

        Parameters
        ----------
        atz: int
            Atomic number of the atom

        Returns
        -------
        List[int]
            List of truncate indices

        """
        nr = _get_nr(self._nr, atz)
        return self._truncate_idxs[nr][atz]

    def _get_truncate_precs(self, atz: int) -> List[int]:
        """Return the truncate precisions for the given atom z

        Parameters
        ----------
        atz: int
            Atomic number of the atom

        Returns
        -------
        List[int]
            List of truncate precisions

        """
        nr = _get_nr(self._nr, atz)
        return self._truncate_precs[nr][atz]

    def to_truncate(self, atz: int) -> bool:
        """Decide whether to truncate the atom's grid

        Parameters
        ----------
        atz: int
            Atomic number of the atom

        Returns
        -------
        bool
            True if the grid should be truncated, False otherwise

        """
        nr = _get_nr(self._nr, atz)
        return atz in self._truncate_idxs[nr]

    def rad_slices(self, atz: int, radgrid: RadialGrid) -> List[slice]:
        """Get the list of slices of radial grid

        Parameters
        ----------
        atz: int
            Atomic number of the atom
        radgrid: RadialGrid
            RadialGrid object of the atom

        Returns
        -------
        List[slice]
            List of slices of the radial grid

        """
        idxs = self._get_truncate_idxs(atz)
        return [slice(idxs[i], idxs[i + 1], None) for i in range(len(idxs) - 1)]

    def precs(self, atz: int, radgrid: RadialGrid) -> List[int]:
        """Get the list of precisions of angular grid for each slice in the
        sliced radial grids

        Parameters
        ----------
        atz: int
            Atomic number of the atom
        radgrid: RadialGrid
            RadialGrid object of the atom

        Returns
        -------
        List[int]
            List of precisions of angular grid for each slice in the
            sliced radial grids

        """
        return self._get_truncate_precs(atz)


class NWChemTrunc(BaseTruncationRules):
    """
    NWChem truncation rules.
    From https://github.com/pyscf/pyscf/blob/18030c75a5c69c1da84574d111693074a622de56/pyscf/dft/gen_grid.py#L122

    Examples
    --------
    >>> from deepchem.utils.dft_utils import NWChemTrunc
    >>> rule = NWChemTrunc([0.5, 1.0, 1.5, 2.0, 2.5, 3.0, 3.5],
    ...                     13, [3, 5, 7, 9, 13],
    ...                     torch.float64, torch.device('cpu'))
    >>> rule.to_truncate(10)
    True

    """

    def __init__(self, radii_list: List[float],
                 prec: Union[int, Callable[[int], int]], precs_list: List[int],
                 dtype: torch.dtype, device: torch.device):
        """Initialize the NWChemTrunc object.

        Parameters
        ----------
        radii_list: List[float]
            List of atomic radii
        prec: Union[int, Callable[[int], int]]
            Precision as a number or a function of atomic number
        precs_list: List[int]
            Complete list of available precision
        dtype: torch.dtype
            Data type of the tensors
        device: torch.device
            Device of the tensors

        Returns
        -------
        NWChemTrunc
            Initialized NWChemTrunc object

        """
        self._radii_list = radii_list
        self._alphas = torch.tensor([
            [0.25, 0.5, 1.0, 4.5],
            [0.1667, 0.5, 0.9, 3.5],
            [0.1, 0.4, 0.8, 2.5],
        ],
                                    dtype=dtype,
                                    device=device)
        self._prec = prec  # precision as a number or a function of atomz
        self._precs_list = precs_list  # complete list of available precision

    def _get_precs(self, atz: int) -> List[int]:
        """Returns the list of precisions

        Parameters
        ----------
        atz: int
            Atomic number of the atom

        Returns
        -------
        List[int]
            List of precisions

        """
        prec_val = _get_nr(self._prec, atz)
        if prec_val == 13:
            precs_idxs = [5, 6, 6, 6, 5]
            res = [self._precs_list[ii] for ii in precs_idxs]
            return res
        elif prec_val >= 13:
            idx: int = self._precs_list.index(prec_val)
            precs_idxs = [5, 7, idx - 1, idx, idx - 1]
            res = [self._precs_list[ii] for ii in precs_idxs]
            return res
        else:
            raise RuntimeError(
                "This shouldn't be displayed. Please report to Github")

    def to_truncate(self, atz: int) -> bool:
        """Decide whether to truncate the atom's grid

        Parameters
        ----------
        atz: int
            Atomic number of the atom

        Returns
        -------
        bool
            True if the grid should be truncated, False otherwise

        """
        prec_val = _get_nr(self._prec, atz)
        if prec_val < 13:
            return False
        return True

    def rad_slices(self, atz: int, radgrid: RadialGrid) -> List[slice]:
        """Get the list of slices of radial grid

        Parameters
        ----------
        atz: int
            Atomic number of the atom
        radgrid: RadialGrid
            RadialGrid object of the atom

        Returns
        -------
        List[slice]
            List of slices of the radial grid

        """
        ratom = self._radii_list[atz]
        ralphas = self._alphas * ratom
        rgrid = radgrid.get_rgrid().reshape(-1, 1)  # (nr, 1)
        if atz <= 2:  # H & He
            ralphas_i = ralphas[0]
        elif atz <= 10:
            ralphas_i = ralphas[1]
        else:
            ralphas_i = ralphas[2]
        # place has value from 0 to 4 (inclusive)
        place = torch.sum(rgrid > ralphas_i, dim=-1)  # (nr,)

        # convert it to slice
        pl, counts = torch.unique_consecutive(place, return_counts=True)
        idx = 0
        res: List[slice] = []
        precs = self._get_precs(atz)
        for i in range(len(precs)):
            c = int(counts[i])
            res.append(slice(idx, idx + c, None))
            idx += c
        return res

    def precs(self, atz: int, radgrid: RadialGrid) -> List[int]:
        """Get the list of precisions of angular grid for each slice in the
        sliced radial grids

        Parameters
        ----------
        atz: int
            Atomic number of the atom
        radgrid: RadialGrid
            RadialGrid object of the atom

        Returns
        -------
        List[int]
            List of precisions of angular grid for each slice in the
            sliced radial grids

        """
        return self._get_precs(atz)


def _get_nr(nr: Union[int, Callable[[int], int]], atz: int) -> int:
    """If nr is a number, return nr, if it is a function, call it with
    atz as the input

    Examples
    --------
    >>> from deepchem.utils.dft_utils.grid.truncation_rules import _get_nr
    >>> _get_nr(5, 1)
    5

    Parameters
    ----------
    nr: Union[int, Callable[[int], int]]
        Number of radial points or a function of atomic number to
        get the number of radial points.
    atz: int
        Atomic number of the atom

    Returns
    -------
    int
        Number of radial points

    """
    if isinstance(nr, int):
        return nr
    else:
        return nr(atz)



import contextlib
import copy
from typing import Mapping, Callable, Union, Dict, List


def set_default_option(defopt: Dict, opt: Dict) -> Dict:
    """return a dictionary based on the options and if no item from option,
    take it from defopt make a shallow copy to detach the results from defopt

    Examples
    --------
    >>> set_default_option({'a': 1, 'b': 2}, {'a': 3})
    {'a': 3, 'b': 2}

    Parameters
    ----------
    defopt: dict
        Default options
    opt: dict
        Options

    Returns
    -------
    dict
        Merged options

    """
    res = copy.copy(defopt)
    res.update(opt)
    return res


def get_and_pop_keys(dct: Dict, keys: List) -> Dict:
    """Get and pop keys from a dictionary

    Examples
    --------
    >>> get_and_pop_keys({'a': 1, 'b': 2}, ['a'])
    {'a': 1}

    Parameters
    ----------
    dct: dict
        Dictionary to pop from
    keys: list
        Keys to pop

    Returns
    -------
    dict
        Dictionary containing the popped keys

    """
    res = {}
    for k in keys:
        res[k] = dct.pop(k)
    return res


def get_method(algname: str, methods: Mapping[str, Callable],
               method: Union[str, Callable]) -> Callable:
    """Get a method from a dictionary of methods

    Examples
    --------
    >>> get_method('foo', {'bar': lambda: 1}, 'bar')()
    1

    Parameters
    ----------
    algname: str
        Name of the algorithm
    methods: dict
        Dictionary of methods
    method: str or callable
        Method to get

    Returns
    -------
    callable
        The method

    """

    if isinstance(method, str):
        methodname = method.lower()
        if methodname in methods:
            return methods[methodname]
        else:
            raise RuntimeError("Unknown %s method: %s" % (algname, method))
    elif hasattr(method, "__call__"):
        return method
    elif method is None:
        assert False, "Internal assert failed, method in get_method is not supposed" \
            "to be None. If this shows, then the corresponding function fails to " \
            "set the default method"
    else:
        raise TypeError(
            "Invalid method type: %s. Only str and callable are accepted." %
            type(method))


@contextlib.contextmanager
def dummy_context_manager():
    """Dummy context manager"""
    yield None


def assert_runtime(cond, msg=""):
    """Assert at runtime

    Examples
    --------
    >>> assert_runtime(False, "This is a test")
    Traceback (most recent call last):
    ...
    RuntimeError: This is a test

    Parameters
    ----------
    cond: bool
        Condition to assert
    msg: str
        Message to raise if condition is not met

    Raises
    ------
    RuntimeError
        If condition is not met

    """
    if not cond:
        raise RuntimeError(msg)



import torch
import logging
from typing import List, Optional
from deepchem.utils import get_memory
from deepchem.utils.differentiation_utils import LinearOperator
from deepchem.utils.dft_utils import DensityFitInfo, OrbitalOrthogonalizer, LibcintWrapper, coul2c, coul3c, overlap, BaseDF

logger = logging.getLogger(__name__)


class DFMol(BaseDF):
    """
    DFMol represents the class of density fitting for an isolated molecule.

    Density fitting is a standard technique in quantum chemistry as it
    helps to accelerate certain parts of a calculation, such as the
    computation of the electron repulsion energy, without significant
    loss of accuracy.

    Examples
    --------
    >>> from deepchem.utils.dft_utils import LibcintWrapper, OrbitalOrthogonalizer, DFMol, DensityFitInfo, AtomCGTOBasis, LibcintWrapper, loadbasis
    >>> import torch
    >>> dtype = torch.double
    >>> d = 1.0
    >>> pos_requires_grad = True
    >>> pos1 = torch.tensor([0.1 * d,  0.0 * d,  0.2 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> pos2 = torch.tensor([0.0 * d,  1.0 * d, -0.4 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> pos3 = torch.tensor([0.2 * d, -1.4 * d, -0.9 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> poss = [pos1, pos2, pos3]
    >>> atomzs = [1, 1, 1]
    >>> allbases = [
    ...     loadbasis("%d:%s" % (max(atomz, 1), "3-21G"), dtype=dtype, requires_grad=False)
    ...     for atomz in atomzs
    ... ]
    >>> atombases = [
    ...     AtomCGTOBasis(atomz=atomzs[i], bases=allbases[i], pos=poss[i])
    ...     for i in range(len(allbases))
    ... ]
    >>> wrapper = LibcintWrapper(atombases, True, None)
    >>> dfinfo = DensityFitInfo("coulomb", atombases)
    >>> orthozer = OrbitalOrthogonalizer(torch.eye(6, dtype=dtype))
    >>> dfmol = DFMol(dfinfo, wrapper, orthozer)
    >>> mol = dfmol.build()
    >>> dm = torch.rand(2, 6, 1)
    >>> elrep = dfmol.get_elrep(dm.to(dtype))
    >>> elrep.fullmatrix().shape
    torch.Size([6, 6])

    """

    def __init__(self,
                 dfinfo: DensityFitInfo,
                 wrapper: LibcintWrapper,
                 orthozer: Optional[OrbitalOrthogonalizer] = None):
        """Initializes the DFMol Class.

        Parameters
        ----------
        dfinfo: DensityFitInfo
            Info about DF Method name and Auxiliary Basis Set.
        wrapper: LibcintWrapper
            Python wrapper for storing info of environment and
            parameters for the integrals calculation.
        orthozer: Optional[OrbitalOrthogonalizer] (default None)
            OrbitalOrthogonalizer for converting orbital to another
            type of orbital by orthogonalizing the basis sets.

        """
        self.dfinfo = dfinfo
        self.wrapper = wrapper
        self._is_built = False
        self._precompute_elmat = True
        self._orthozer = orthozer

    def build(self) -> BaseDF:
        """Build the Density Fitting Object

        Returns
        -------
        BaseDF
            The constructed density fitting object.

        """
        self._is_built = True

        # construct the matrix used to calculate the electron repulsion for
        # density fitting method
        method = self.dfinfo.method
        auxbasiswrapper = LibcintWrapper(self.dfinfo.auxbasis,
                                         spherical=self.wrapper.spherical)
        basisw, auxbw = LibcintWrapper.concatenate(self.wrapper,
                                                   auxbasiswrapper)

        if method == "coulomb":
            logger.info("Calculating the 2e2c integrals")
            j2c = coul2c(auxbw)  # (nxao, nxao)
            logger.info("Calculating the 2e3c integrals")
            j3c = coul3c(basisw, other1=basisw,
                         other2=auxbw)  # (nao, nao, nxao)
        elif method == "overlap":
            j2c = overlap(auxbw)  # (nxao, nxao)
            # TODO: implement overlap3c
            raise NotImplementedError(
                "Density fitting with overlap minimization is not implemented")
        self._j2c = j2c  # (nxao, nxao)
        self._j3c = j3c  # (nao, nao, nxao)
        logger.info("Precompute matrix for density fittings")
        self._inv_j2c = torch.inverse(j2c)

        # if the memory is too big, then don't precompute elmat
        if get_memory(j3c) > 10 * 1024**3:
            self._precompute_elmat = False
        else:
            self._precompute_elmat = True
            self._el_mat = torch.matmul(j3c, self._inv_j2c)  # (nao, nao, nxao)

        logger.info("Density fitting done")
        return self

    def get_elrep(self, dm: torch.Tensor) -> LinearOperator:
        """
        Construct the electron repulsion linear operator from the given density
        matrix using the density fitting method.

        Parameters
        ----------
        dm : torch.Tensor
            Density matrix. (*BD, nao, nao)

        Returns
        -------
        LinearOperator
            The electron repulsion linear operator.

        """

        # convert the dm into the original cgto basis
        if self._orthozer is not None:
            dm = self._orthozer.unconvert_dm(dm)

        if self._precompute_elmat:
            df_coeffs = torch.einsum("...ij,ijk->...k", dm,
                                     self._el_mat)  # (*BD, nxao)
        else:
            temp = torch.einsum("...ij,ijl->...l", dm, self._j3c)
            df_coeffs = torch.einsum("...l,lk->...k", temp,
                                     self._inv_j2c)  # (*BD, nxao)

        mat = torch.einsum("...k,ijk->...ij", df_coeffs,
                           self._j3c)  # (*BD, nao, nao)
        mat = (mat + mat.transpose(-2, -1)) * 0.5
        if self._orthozer is not None:
            mat = self._orthozer.convert2(mat)
        return LinearOperator.m(mat, is_hermitian=True)

    @property
    def j2c(self) -> torch.Tensor:
        """Returns the 2-centre 2-electron integrals of the auxiliary basis.

        Returns
        -------
        torch.Tensor
            2-centre 2-electron integrals of the auxiliary basis.

        """
        return self._j2c

    @property
    def j3c(self) -> torch.Tensor:
        """
        Return the 3-centre 2-electron integrals of the auxiliary basis and the
        basis.

        Returns
        -------
        torch.Tensor
            3-centre 2-electron integrals of the auxiliary basis and the basis.

        """
        return self._j3c

    def getparamnames(self, methodname: str, prefix: str = "") -> List[str]:
        """
        This method should list tensor names that affect the output of the
        method with name indicated in ``methodname``.

        Parameters
        ---------
        methodname: str
            The name of the method of the class.
        prefix: str (default="")
            The prefix to be appended in front of the parameters name.
            This usually contains the dots.

        Returns
        -------
        List[str]
            Sequence of name of parameters affecting the output of the method.

        """
        if methodname == "get_elrep":
            if self._precompute_elmat:
                params = [prefix + "_el_mat", prefix + "_j3c"]
            else:
                params = [prefix + "_inv_j2c", prefix + "_j3c"]
            if self._orthozer is not None:
                pfix = prefix + "_orthozer."
                params += self._orthozer.getparamnames("unconvert_dm", prefix=pfix) + \
                    self._orthozer.getparamnames("convert2", prefix=pfix)
            return params
        else:
            raise KeyError("getparamnames has no %s method" % methodname)



"""
HOPV dataset loader.
"""
import os
import deepchem as dc
from deepchem.molnet.load_function.molnet_loader import TransformerGenerator, _MolnetLoader
from deepchem.data import Dataset
from typing import List, Optional, Tuple, Union

HOPV_URL = "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/hopv.tar.gz"
HOPV_TASKS = [
    'HOMO', 'LUMO', 'electrochemical_gap', 'optical_gap', 'PCE', 'V_OC', 'J_SC',
    'fill_factor'
]


class _HOPVLoader(_MolnetLoader):

    def create_dataset(self) -> Dataset:
        dataset_file = os.path.join(self.data_dir, "hopv.csv")
        if not os.path.exists(dataset_file):
            dc.utils.data_utils.download_url(url=HOPV_URL,
                                             dest_dir=self.data_dir)
            dc.utils.data_utils.untargz_file(
                os.path.join(self.data_dir, 'hopv.tar.gz'), self.data_dir)
        loader = dc.data.CSVLoader(tasks=self.tasks,
                                   feature_field="smiles",
                                   featurizer=self.featurizer)
        return loader.create_dataset(dataset_file, shard_size=8192)


def load_hopv(
    featurizer: Union[dc.feat.Featurizer, str] = 'ECFP',
    splitter: Union[dc.splits.Splitter, str, None] = 'scaffold',
    transformers: List[Union[TransformerGenerator, str]] = ['normalization'],
    reload: bool = True,
    data_dir: Optional[str] = None,
    save_dir: Optional[str] = None,
    **kwargs
) -> Tuple[List[str], Tuple[Dataset, ...], List[dc.trans.Transformer]]:
    """Load HOPV datasets. Does not do train/test split

    The HOPV datasets consist of the "Harvard Organic
    Photovoltaic Dataset. This dataset includes 350 small
    molecules and polymers that were utilized as p-type materials
    in OPVs. Experimental properties include: HOMO [a.u.], LUMO
    [a.u.], Electrochemical gap [a.u.], Optical gap [a.u.], Power
    conversion efficiency [%], Open circuit potential [V], Short
    circuit current density [mA/cm^2], and fill factor [%].
    Theoretical calculations in the original dataset have been
    removed (for now).

    Lopez, Steven A., et al. "The Harvard organic photovoltaic dataset." Scientific data 3.1 (2016): 1-7.

    Parameters
    ----------
    featurizer: Featurizer or str
        the featurizer to use for processing the data.  Alternatively you can pass
        one of the names from dc.molnet.featurizers as a shortcut.
    splitter: Splitter or str
        the splitter to use for splitting the data into training, validation, and
        test sets.  Alternatively you can pass one of the names from
        dc.molnet.splitters as a shortcut.  If this is None, all the data
        will be included in a single dataset.
    transformers: list of TransformerGenerators or strings
        the Transformers to apply to the data.  Each one is specified by a
        TransformerGenerator or, as a shortcut, one of the names from
        dc.molnet.transformers.
    reload: bool
        if True, the first call for a particular featurizer and splitter will cache
        the datasets to disk, and subsequent calls will reload the cached datasets.
    data_dir: str
        a directory to save the raw data in
    save_dir: str
        a directory to save the dataset in
    """
    loader = _HOPVLoader(featurizer, splitter, transformers, HOPV_TASKS,
                         data_dir, save_dir, **kwargs)
    return loader.load_dataset('hopv', reload)



from typing import Any, Callable, Union
import torch
from abc import abstractmethod
from deepchem.utils.differentiation_utils.grad import jac

# taking most of the part from SciPy


class Jacobian(object):
    """Base class for the Jacobians used in rootfinder algorithms.

    A Jacobian can best be defined as a determinant which is defined
    for a finite number of functions of the same number of variables
    in which each row consists of the first partial derivatives of
    the same function with respect to each of the variables.

    References
    ----------
    [1].. https://en.wikipedia.org/wiki/Jacobian_matrix_and_determinant
    [2].. Kasim, Muhammad & Vinko, Sam. (2020). xi$-torch: differentiable scientific computing library.

    """

    @abstractmethod
    def setup(self, x0: torch.Tensor, y0: torch.Tensor, func: Callable):
        """Setup the Jacobian for the rootfinder."""
        pass

    @abstractmethod
    def solve(self, v: torch.Tensor, tol: Any = 0):
        """Solve the linear system `J dx = v`."""
        pass

    @abstractmethod
    def update(self, x: torch.Tensor, y: torch.Tensor):
        """Update the Jacobian approximation."""
        pass


class BroydenFirst(Jacobian):
    """
    Approximating the Jacobian based on Broyden's first approximation.

    Examples
    --------
    >>> from deepchem.utils.differentiation_utils.optimize.jacobian import BroydenFirst
    >>> jacobian = BroydenFirst()
    >>> x0 = torch.tensor([1.0, 1.0], requires_grad=True)
    >>> def func(x):
    ...     return torch.tensor([x[0]**2 + x[1]**2 - 1.0, x[0] - x[1]])
    >>> y0 = func(x0)
    >>> v = torch.tensor([1.0, 1.0])
    >>> jacobian.setup(x0, y0, func)
    >>> jacobian.solve(v)
    tensor([-0.7071, -0.7071], grad_fn=<MulBackward0>)

    References
    ----------
    [1].. B.A. van der Rotten, PhD thesis,
        "A limited memory Broyden method to solve high-dimensional
        systems of nonlinear equations". Mathematisch Instituut,
        Universiteit Leiden, The Netherlands (2003).

    """

    def __init__(self,
                 alpha: Union[torch.Tensor, None] = None,
                 uv0: Any = None,
                 max_rank: Union[float, None] = None):
        """The initial guess of inverse Jacobian is `-alpha * I + u v^T`.
        `max_rank` indicates the maximum rank of the Jacoabian before
        reducing it

        Parameters
        ----------
        alpha : Union[torch.Tensor, None]
            The initial guess of inverse Jacobian is `-alpha * I`.
            If None, it is set to `-1.0`.
        uv0 : tuple, optional
            The initial guess of the inverse Jacobian.
        max_rank : Union[float, None]
            The maximum rank of the Jacobian before reducing it.
            If None, it is set to `inf`.

        """
        self.alpha = alpha
        self.uv0 = uv0
        self.max_rank = max_rank

    def setup(self, x0: torch.Tensor, y0: torch.Tensor, func: Callable):
        """Setup the Jacobian for the rootfinder.

        Parameters
        ----------
        x0
            The initial guess of the root.
        y0
            The function value at the initial guess.
        func
            The function to find the root.

        """
        self.x_prev = x0
        self.y_prev = y0

        if self.max_rank is None:
            self.max_rank = float('inf')

        if self.alpha is None:
            normy0 = torch.norm(y0)
            ones = torch.ones_like(normy0)
            if normy0:
                self.alpha = 0.5 * torch.max(torch.norm(x0), ones) / normy0
            else:
                self.alpha = ones

        if self.uv0 == "svd":
            self.uv0 = _get_svd_uv0(func, x0)

        # setup the approximate inverse Jacobian
        self.Gm = LowRankMatrix(-self.alpha, self.uv0, "restart")

    def _reduce(self):
        """
        reduce the size of Gm
        initially it was a lambda function, but it causes a leak, so
        I arranged into a method to remove the leak
        """
        self.Gm.reduce(self.max_rank)

    def solve(self, v: torch.Tensor, tol=0) -> torch.Tensor:
        """Solve the linear system `J dx = v`.

        Parameters
        ----------
        v: torch.Tensor
            The right-hand side of the linear system.
        tol: torch.Tensor
            The tolerance for the linear system.

        Returns
        -------
        res: torch.Tensor
            The solution of the linear system.

        """
        res = self.Gm.mv(v)
        return res

    def update(self, x: torch.Tensor, y: torch.Tensor):
        """Update the Jacobian approximation.

        Parameters
        ----------
        x: torch.Tensor
            The current point.
        y: torch.Tensor
            The function value at the current point.

        """
        dy = y - self.y_prev
        dx = x - self.x_prev
        # update Gm
        self._update(x, y, dx, dy, dx.norm(), dy.norm())

        self.y_prev = y
        self.x_prev = x

    def _update(self, x: torch.Tensor, y: torch.Tensor, dx: torch.Tensor,
                dy: torch.Tensor, dxnorm: torch.Tensor, dynorm: torch.Tensor):
        """Update the Jacobian approximation.

        Parameters
        ----------
        x: torch.Tensor
            The current point.
        y: torch.Tensor
            The function value at the current point.
        dx: torch.Tensor
            The difference between the current point and the previous point.
        dy: torch.Tensor
            The difference between the function value at the current point
            and the previous point.
        dxnorm: torch.Tensor
            The norm of `dx`.
        dynorm: torch.Tensor
            The norm of `dy`.

        """
        # keep the rank small
        self._reduce()

        v = self.Gm.rmv(dx)
        c = dx - self.Gm.mv(dy)
        d = v / torch.dot(dy, v)
        self.Gm = self.Gm.append(c, d)


class BroydenSecond(BroydenFirst):
    """
    Inverse Jacobian approximation based on Broyden's second method.

    Examples
    --------
    >>> from deepchem.utils.differentiation_utils.optimize.jacobian import BroydenSecond
    >>> jacobian = BroydenSecond()
    >>> x0 = torch.tensor([1.0, 1.0], requires_grad=True)
    >>> def func(x):
    ...     return torch.tensor([x[0]**2 + x[1]**2 - 1.0, x[0] - x[1]])
    >>> y0 = func(x0)
    >>> v = torch.tensor([1.0, 1.0])
    >>> jacobian.setup(x0, y0, func)
    >>> jacobian.solve(v)
    tensor([-0.7071, -0.7071], grad_fn=<MulBackward0>)

    References
    ----------
    [1] B.A. van der Rotten, PhD thesis,
        "A limited memory Broyden method to solve high-dimensional
        systems of nonlinear equations". Mathematisch Instituut,
        Universiteit Leiden, The Netherlands (2003).

    """

    def _update(self, x: torch.Tensor, y: torch.Tensor, dx: torch.Tensor,
                dy: torch.Tensor, dxnorm: torch.Tensor, dynorm: torch.Tensor):
        """Update the Jacobian approximation.

        Parameters
        ----------
        x: torch.Tensor
            The current point.
        y: torch.Tensor
            The function value at the current point.
        dx: torch.Tensor
            The difference between the current point and the previous point.
        dy: torch.Tensor
            The difference between the function value at the current point
            and the previous point.
        dxnorm: torch.Tensor
            The norm of `dx`.
        dynorm: torch.Tensor
            The norm of `dy`.

        """
        # keep the rank small
        self._reduce()

        v = dy
        c = dx - self.Gm.mv(dy)
        d = v / (dynorm * dynorm)
        self.Gm = self.Gm.append(c, d)


class LinearMixing(Jacobian):
    """ Approximating the Jacobian based on linear mixing.
    It acts as a simple check for the functionality of the rootfinder.

    Examples
    --------
    >>> from deepchem.utils.differentiation_utils.optimize.jacobian import LinearMixing
    >>> jacobian = LinearMixing()
    >>> x0 = torch.tensor([1.0, 1.0], requires_grad=True)
    >>> def func(x):
    ...     return torch.tensor([x[0]**2 + x[1]**2 - 1.0, x[0] - x[1]])
    >>> y0 = func(x0)
    >>> v = torch.tensor([1.0, 1.0])
    >>> jacobian.setup(x0, y0, func)
    >>> jacobian.solve(v)
    tensor([1., 1.])

    """

    def __init__(self, alpha: Union[float, None] = None):
        """The initial guess of inverse Jacobian is ``-alpha * I``

        Parameters
        ----------
        alpha : float, optional
            The initial guess of inverse Jacobian is ``-alpha * I``.
            If None, it is set to ``-1.0``.

        """
        if alpha is None:
            alpha = -1.0
        self.alpha = alpha

    def setup(self, x0: torch.Tensor, y0: torch.Tensor, func: Callable):
        """Setup the Jacobian for the rootfinder.

        Parameters
        ----------
        x0: torch.Tensor
            The initial guess of the root.
        y0: torch.Tensor
            The function value at the initial guess.
        func: Callable
            The function to find the root.

        """
        pass

    def solve(self, v: torch.Tensor, tol=0) -> torch.Tensor:
        """Solve the linear system `J dx = v`.

        Parameters
        ----------
        v: torch.Tensor
            The right-hand side of the linear system.
        tol
            The tolerance for the linear system.

        """
        return -v * self.alpha

    def update(self, x: torch.Tensor, y: torch.Tensor):
        """Update the Jacobian approximation.

        Parameters
        ----------
        x: torch.Tensor
            The current point.
        y: torch.Tensor
            The function value at the current point.

        """
        pass


class LowRankMatrix(object):
    """represents a matrix of `\alpha * I + \sum_n c_n d_n^T`

    Examples
    --------
    >>> from deepchem.utils.differentiation_utils.optimize.jacobian import LowRankMatrix
    >>> import torch
    >>> alpha = 1.0
    >>> uv0 = (torch.tensor([1.0, 1.0]), torch.tensor([1.0, 1.0]))
    >>> reduce_method = "restart"
    >>> matrix = LowRankMatrix(alpha, uv0, reduce_method)
    >>> v = torch.tensor([1.0, 1.0])
    >>> matrix.mv(v)
    tensor([3., 3.])
    >>> matrix.rmv(v)
    tensor([3., 3.])

    """

    def __init__(self, alpha: torch.Tensor, uv0, reduce_method: str):
        """initialize the matrix

        Parameters
        ----------
        alpha : torch.Tensor
            The coefficient of the identity matrix
        uv0 : tuple
            The initial guess of the inverse Jacobian
        reduce_method : str
            The method to reduce the rank of the matrix

        """
        self.alpha = alpha
        if uv0 is None:
            self.cns = []
            self.dns = []
        else:
            cn0, dn0 = uv0
            self.cns = [cn0]
            self.dns = [dn0]
        self.reduce_method = {"restart": 0, "simple": 1}[reduce_method]

    def mv(self, v: torch.Tensor) -> torch.Tensor:
        """multiply the matrix with a vector

        Parameters
        ----------
        v: torch.Tensor
            Vector to multiply

        Returns
        -------
        res: torch.Tensor
            Result of the multiplication

        """
        res = self.alpha * v
        for i in range(len(self.dns)):
            res += self.cns[i] * torch.dot(self.dns[i], v)
        return res

    def rmv(self, v: torch.Tensor) -> torch.Tensor:
        """multiply the transpose of the matrix with a vector

        Parameters
        ----------
        v: torch.Tensor
            Vector to multiply

        Returns
        -------
        res: torch.Tensor
            Result of the multiplication

        """
        res = self.alpha * v
        for i in range(len(self.dns)):
            res += self.dns[i] * torch.dot(self.cns[i], v)
        return res

    def append(self, c: torch.Tensor, d: torch.Tensor):
        """append a rank-1 matrix to the matrix

        Parameters
        ----------
        c: torch.Tensor
            The first vector
        d: torch.Tensor
            The second vector

        Returns
        -------
        res: Union['LowRankMatrix', 'FullRankMatrix']
            The matrix after appending the rank-1 matrix

        """
        self.cns.append(c)
        self.dns.append(d)
        if len(self.cns) >= torch.numel(c):
            return FullRankMatrix(self.alpha, self.cns, self.dns)
        return self

    def reduce(self, max_rank: int, **otherparams):
        """reduce the rank of the matrix

        Parameters
        ----------
        max_rank : int
            The maximum rank of the matrix
        otherparams
            Other parameters

        """
        if len(self.cns) > max_rank:
            if self.reduce_method == 0:  # restart
                del self.cns[:]
                del self.dns[:]
            elif self.reduce_method == 1:  # simple
                n = len(self.cns)
                del self.cns[:n - max_rank]
                del self.dns[:n - max_rank]


class FullRankMatrix(object):
    """represents a full rank matrix of `\alpha * I + \sum_n c_n d_n^T`

    Examples
    --------
    >>> from deepchem.utils.differentiation_utils.optimize.jacobian import FullRankMatrix
    >>> import torch
    >>> alpha = 1.0
    >>> cns = [torch.tensor([1.0, 1.0]), torch.tensor([1.0, 1.0])]
    >>> dns = [torch.tensor([1.0, 1.0]), torch.tensor([1.0, 1.0])]
    >>> matrix = FullRankMatrix(alpha, cns, dns)
    >>> v = torch.tensor([1.0, 1.0])
    >>> matrix.mv(v)
    tensor([5., 5.])
    >>> matrix.rmv(v)
    tensor([5., 5.])

    """

    def __init__(self, alpha: torch.Tensor, cns: Any, dns: Any):
        """initialize the matrix

        Parameters
        ----------
        alpha : torch.Tensor
            Coefficient of the identity matrix
        cns : List
            List of the first vectors
        dns : List
            List of the second vectors

        """
        size = torch.numel(cns[0])
        dtype, device = cns[0].dtype, cns[0].device
        self.mat = torch.eye(size, dtype=dtype, device=device)
        self.mat *= alpha
        for i in range(len(cns)):
            self.mat += torch.ger(cns[i], dns[i])

    def mv(self, v: torch.Tensor) -> torch.Tensor:
        """multiply the matrix with a vector

        Parameters
        ----------
        v: torch.Tensor
            The vector to multiply

        Returns
        -------
        res: torch.Tensor
            The result of the multiplication

        """
        res = torch.matmul(self.mat, v)
        return res

    def rmv(self, v: torch.Tensor) -> torch.Tensor:
        """multiply the transpose of the matrix with a vector

        Parameters
        ----------
        v: torch.Tensor
            The vector to multiply

        Returns
        -------
        torch.Tensor
            The result of the multiplication

        """
        return torch.matmul(self.mat.T, v)

    def append(self, c: torch.Tensor, d: torch.Tensor):
        """append a rank-1 matrix to the matrix

        Parameters
        ----------
        c: torch.Tensor
            The first vector
        d: torch.Tensor
            The second vector

        Returns
        -------
        FullRankMatrix
            The matrix after appending the rank-1 matrix

        """
        self.mat += torch.ger(c, d)
        return self

    def reduce(self, max_rank: int, **kwargs):
        """reduce the rank of the matrix

        Parameters
        ----------
        max_rank : int
            The maximum rank of the matrix
        otherparams
            Other parameters

        """
        pass


def _get_svd_uv0(func: Callable, x0: torch.Tensor) -> tuple:
    """get the initial guess of the inverse Jacobian from the first Jacobian

    Parameters
    ----------
    func: Callable
        The function to find the root
    x0: torch.Tensor
        The initial guess of the root

    Returns
    -------
    uv0: tuple
        The initial guess of the inverse Jacobian

    """
    from deepchem.utils.differentiation_utils import svd
    # raise RuntimeError
    fjac = jac(func, (x0.clone().requires_grad_(),), idxs=[0])[0]
    # u: (n, 1), s: (1,), vh: (1, n)
    u, s, vh = svd(fjac, k=1, mode="lowest", method="davidson", min_eps=1e-3)
    sinv_sqrt = 1. / torch.sqrt(torch.clamp(s, min=0.1))
    uv0 = (sinv_sqrt * vh.squeeze(-2), sinv_sqrt * u.squeeze(-1))
    return uv0
